from io import BytesIO

from django.core.cache import cache
from django.core.files.uploadedfile import InMemoryUploadedFile

from .constants import CACHE_TIMEOUT
from .utils import log


class FileCache:
    """Cache file data and retain the file upon confirmation."""
    timeout = CACHE_TIMEOUT

    def __init__(self):
        self.cache = cache
        self.cached_keys = []

    def set(self, key, upload):
        """
        Set file data to cache for 1000s

        :param key: cache key
        :param upload: file data
        """
        try:  # noqa: WPS229
            state = {
                'name': upload.name,
                'size': upload.size,

                'charset': upload.charset,
                'content': upload.file.read(),

                'content_type': upload.content_type,
            }

            upload.file.seek(0)
            self.cache.set(key, state, self.timeout)

            log(f'Setting file cache with {key}')
            self.cached_keys.append(key)
        except AttributeError:  # pragma: no cover
            pass  # noqa: WPS420

    def get(self, key):
        """
        Get the file data from cache using specific cache key

        :param key: cache key
        :return: File data
        """
        upload = None
        state = self.cache.get(key)

        if state:
            file = BytesIO()
            file.write(state['content'])

            upload = InMemoryUploadedFile(
                content_type=state['content_type'],
                charset=state['charset'],
                field_name='file',
                name=state['name'],
                size=state['size'],
                file=file,
            )

            upload.file.seek(0)
            log(f'Getting file cache with {key}')

        return upload

    def delete(self, key):
        """
        Delete file data from cache

        :param key: cache key
        """
        self.cache.delete(key)
        self.cached_keys.remove(key)

    def delete_all(self):
        """Delete all cached file data from cache."""
        self.cache.delete_many(self.cached_keys)
        self.cached_keys = []



from django import template
from django.utils.html import escape
from django.utils.safestring import mark_safe

register = template.Library()


@register.filter
def format_change_data_field_value(field_value):
    if isinstance(field_value, str):
        return field_value

    try:
        output = '<ul>'

        for value in iter(field_value):
            output += '<li>' + escape(value) + '</li>'

        output += '</ul>'

        return mark_safe(output)
    except Exception:  # noqa
        return field_value


@register.simple_tag
def verbose_name(obj, field_name):
    if obj:
        return obj._meta.get_field(field_name).verbose_name  # noqa



import itertools
import warnings

from django import template
from django.contrib.admin.utils import lookup_field
from django.core.exceptions import ObjectDoesNotExist
from django.db.models import ForeignKey
from django.template.defaulttags import NowNode
from django.urls import NoReverseMatch
from django.urls import reverse
from django.utils.safestring import mark_safe

from .. import utils
from ..config import get_config

django_version = utils.django_major_version()

register = template.Library()
simple_tag = register.simple_tag





@register.filter(name='elegant_conf')
def elegant_conf(name):
    value = get_config(name)
    return mark_safe(value) if isinstance(value, str) else value


@register.tag
def elegant_date(parser, token):
    return NowNode(get_config('HEADER_DATE_FORMAT'))


@register.tag
def elegant_time(parser, token):
    return NowNode(get_config('HEADER_TIME_FORMAT'))


@register.filter
def field_contents_foreign_linked(admin_field):
    """Return the .contents attribute of the admin_field, and if it
    is a foreign key, wrap it in a link to the admin page for that
    object.

    Use by replacing '{{ field.contents }}' in an admin template (e.g.
    fieldset.html) with '{{ field|field_contents_foreign_linked }}'.
    """
    # print(admin_field.form.Meta.fields)

    fieldname = admin_field.field['field']
    displayed = admin_field.contents()
    obj = admin_field.form.instance

    if django_version > (3, 1):
        return displayed

    if not hasattr(admin_field.model_admin, 'linked_readonly_fields'):
        return displayed

    if fieldname not in admin_field.model_admin.linked_readonly_fields:
        return displayed

    # if admin_field.is_readonly:
    #     return displayed

    try:
        fieldtype, attr, value = lookup_field(fieldname, obj, admin_field.model_admin)
    except ObjectDoesNotExist:
        fieldtype = None
        value = None

    if isinstance(fieldtype, ForeignKey):
        try:
            url = admin_url(value)
        except NoReverseMatch:
            url = None

        if url:
            displayed = f"<a href='{url}'>{displayed}</a>"

    return mark_safe(displayed)


@register.filter
def admin_url(obj):
    info = (obj._meta.app_label, obj._meta.object_name.lower())
    return reverse('admin:%s_%s_change' % info, args=[obj.pk])


@register.simple_tag
def elegant_bc(*args):
    return utils.value_by_version(args)


@simple_tag
def elegant_bc_value(*args):
    return utils.value_by_version(args)


@simple_tag
def admin_extra_filters(cl):
    """ Return the dict of used filters which is not included
    in list_filters form """
    used_parameters = list(itertools.chain(*(s.used_parameters.keys() for s in cl.filter_specs)))
    return {k: v for k, v in cl.params.items() if k not in used_parameters}


@simple_tag
def elegant_django_version():
    return django_version


@register.filter
def django_version_lt(string):
    return django_version < str_to_version(string)


@register.filter
def django_version_lte(string):
    return django_version <= str_to_version(string)


@register.filter
def django_version_gt(string):
    return django_version > str_to_version(string)


@register.filter
def django_version_gte(string):
    return django_version >= str_to_version(string)


def str_to_version(string):
    return tuple([int(s) for s in string.split('.')])

# if django_version < (1, 9):
#     # Add empty tags to avoid Django template errors if < Django 1.9
#     @register.simple_tag
#     def add_preserved_filters(*args, **kwargs):
#         pass

# if django_version < (1, 5):
#     # Add admin_urlquote filter to support Django 1.4
#     from django.contrib.admin.util import quote
#
#
#     @register.filter
#     def admin_urlquote(value):
#         return quote(value)



# flake8: noqa
from .mixins.objects import action
from .mixins.objects import BaseDjangoObjectActions
from .mixins.objects import DjangoObjectActions
from .mixins.objects import takes_instance_or_queryset

__version__ = '0.2.6'
VERSION = __version__



from django.conf import settings

SAVE = '_save'

SAVE_AS_NEW = '_saveasnew'
ADD_ANOTHER = '_addanother'

SAVE_AND_CONTINUE = '_continue'
SAVE_ACTIONS = [SAVE, SAVE_AS_NEW, ADD_ANOTHER, SAVE_AND_CONTINUE]

CONFIRM_ADD = '_confirm_add'
CONFIRM_CHANGE = '_confirm_change'
CONFIRMATION_RECEIVED = '_confirmation_received'

CACHE_TIMEOUT = getattr(settings, 'ADMIN_CONFIRM_CACHE_TIMEOUT', 1000)

CACHE_KEYS = {
    'object': 'admin_confirm__confirmation_object',
    'post': 'admin_confirm__confirmation_request_post',
}

CACHE_KEY_PREFIX = getattr(settings, 'ADMIN_CONFIRM_CACHE_KEY_PREFIX', 'admin_confirm__file_cache')

DEBUG = getattr(settings, 'ADMIN_CONFIRM_DEBUG', False)



import copy

from django.conf import settings
from django.contrib import admin
from django.contrib.admin import ModelAdmin
from django.contrib.admin.views.main import ChangeList
from django.db import models
from django.forms import ModelForm

from .compat import ct_admin
from .widgets import ElegantSplitDateTimeWidget
from .widgets import NumberInput


class SortableModelAdminBase:
    """
    Base class for SortableTabularInline and SortableModelAdmin
    """
    sortable = 'order'

    class Media:
        js = ('elegant/js/sortables.js',)


class SortableListForm(ModelForm):
    """
    Just Meta holder class
    """

    class Meta:
        widgets = {
            'order': NumberInput(attrs={'class': 'hide input-mini elegant-sortable'})
        }


class SortableChangeList(ChangeList):
    """
    Class that forces ordering by sortable param only
    """

    def get_ordering(self, request, queryset):
        return [self.model_admin.sortable, '-' + self.model._meta.pk.name]  # noqa


class SortableTabularInlineBase(SortableModelAdminBase):
    """
    Sortable tabular inline
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.ordering = (self.sortable,)
        self.fields = self.fields or []  # noqa

        if self.fields and self.sortable not in self.fields:
            self.fields = list(self.fields) + [self.sortable]

    def formfield_for_dbfield(self, db_field, **kwargs):
        if db_field.name == self.sortable:
            kwargs['widget'] = SortableListForm.Meta.widgets['order']

        return super().formfield_for_dbfield(db_field, **kwargs)  # noqa


class SortableTabularInline(SortableTabularInlineBase, admin.TabularInline):
    pass


class SortableGenericTabularInline(SortableTabularInlineBase, ct_admin.GenericTabularInline):
    pass


class SortableStackedInlineBase(SortableModelAdminBase):
    """
    Sortable stacked inline
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.ordering = (self.sortable,)

    def get_fieldsets(self, *args, **kwargs):
        """
        Iterate all fieldsets and make sure sortable is in the first fieldset
        Remove sortable from every other fieldset, if by some reason someone
        has added it
        """
        fieldsets = super().get_fieldsets(*args, **kwargs)  # noqa
        sortable_added = False

        for fieldset in fieldsets:
            for line in fieldset:
                if not line or not isinstance(line, dict):
                    continue

                fields = line.get('fields')

                # Some use tuples for fields however they are immutable
                if isinstance(fields, tuple):
                    raise AssertionError(
                        'The fields attribute of your Inline is a tuple. '
                        'This must be list as we may need to modify it and '
                        'tuples are immutable.')

                if self.sortable in fields:
                    fields.remove(self.sortable)  # noqa

                # Add sortable field always as first
                if not sortable_added:
                    fields.insert(0, self.sortable)  # noqa
                    sortable_added = True
                    break

        return fieldsets

    def formfield_for_dbfield(self, db_field, **kwargs):
        if db_field.name == self.sortable:
            kwargs['widget'] = copy.deepcopy(SortableListForm.Meta.widgets['order'])
            kwargs['widget'].attrs['class'] += ' elegant-sortable-stacked'
            kwargs['widget'].attrs['rowclass'] = ' elegant-sortable-stacked-row'

        return super().formfield_for_dbfield(db_field, **kwargs)  # noqa


class SortableStackedInline(SortableStackedInlineBase, admin.StackedInline):
    pass


class SortableGenericStackedInline(SortableStackedInlineBase, ct_admin.GenericStackedInline):
    pass


class SortableModelAdmin(SortableModelAdminBase, ModelAdmin):
    """
    Sortable tabular inline
    """
    list_per_page = 500

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.ordering = (self.sortable,)

        if self.list_display and self.sortable not in self.list_display:
            self.list_display = list(self.list_display) + [self.sortable]

        self.list_editable = self.list_editable or []

        if self.sortable not in self.list_editable:
            self.list_editable = list(self.list_editable) + [self.sortable]

        self.exclude = self.exclude or []

        if self.sortable not in self.exclude:
            self.exclude = list(self.exclude) + [self.sortable]

    def merge_form_meta(self, form):
        """
        Prepare Meta class with order field widget
        """
        if not getattr(form, 'Meta', None):
            form.Meta = SortableListForm.Meta

        if not getattr(form.Meta, 'widgets', None):
            form.Meta.widgets = {}

        form.Meta.widgets[self.sortable] = SortableListForm.Meta.widgets['order']

    def get_changelist_form(self, request, **kwargs):
        form = super().get_changelist_form(request, **kwargs)
        self.merge_form_meta(form)

        return form

    def get_changelist(self, request, **kwargs):
        return SortableChangeList

    def save_model(self, request, obj, form, change):
        if not obj.pk:
            max_order = obj.__class__.objects.aggregate(models.Max(self.sortable))

            try:
                next_order = max_order['%s__max' % self.sortable] + 1
            except TypeError:
                next_order = 1

            setattr(obj, self.sortable, next_order)

        super().save_model(request, obj, form, change)


# Quite aggressive detection and intrusion into Django CMS
# Didn't found any other solutions though
if 'cms' in settings.INSTALLED_APPS:
    try:
        from cms.admin.forms import PageForm  # noqa

        PageForm.Meta.widgets = {
            'publication_date': ElegantSplitDateTimeWidget,
            'publication_end_date': ElegantSplitDateTimeWidget,
        }
    except ImportError:
        pass


class SortableAdmin(admin.ModelAdmin):
    class Media:
        js = (
            'elegant/js/jquery-ui-1.10.3.sortable.min.js',
            'elegant/js/sortable.changelist.js',
        )


class SortableTabularInline(admin.TabularInline):
    class Media:
        js = (
            'elegant/js/jquery-ui-1.10.3.sortable.min.js',
            'elegant/js/sortable.tabular.inline.js',
        )


class SortableStackedInline(admin.StackedInline):
    class Media:
        js = (
            'elegant/js/jquery-ui-1.10.3.sortable.min.js',
            'elegant/js/sortable.stacked.inline.js',
        )



import weakref

from django.template.loader import render_to_string


class BaseComponent:
    template = None
    instances = []

    def __init__(self, **kwargs):
        self.__class__.instances.append(weakref.proxy(self))
        self.request = kwargs.pop('request')

        self.context = kwargs
        self.context['dom_id'] = self.get_unique_id()

    @classmethod
    def get_unique_id(cls):
        return f'{cls.__name__.lower()}-{len(cls.instances)}'

    def render(self):
        return render_to_string(self.template, self.context, request=self.request)

    def __unicode__(self):
        return self.render()


class Dropdown(BaseComponent):
    template = 'elegant/actions/dropdown.html'

    def __init__(self, **kwargs):
        super().__init__(**kwargs)



from django.conf import settings


def show_notice(request):
    return request.user.is_authenticated or getattr(settings, 'ENVIRONMENT_SHOW_TO_UNAUTHENTICATED', True)


def from_settings(request):
    return {
        'ENVIRONMENT_NAME': getattr(settings, 'ENVIRONMENT_NAME', None),
        'ENVIRONMENT_COLOR': getattr(settings, 'ENVIRONMENT_COLOR', None),
        'ENVIRONMENT_TEXT_COLOR': getattr(settings, 'ENVIRONMENT_TEXT_COLOR', 'white'),
        'ENVIRONMENT_ADMIN_SELECTOR': getattr(settings, 'ENVIRONMENT_ADMIN_SELECTOR', 'body'),
        'ENVIRONMENT_FLOAT': getattr(settings, 'ENVIRONMENT_FLOAT', False),
        'show_notice': show_notice(request),

    }



from django import template
from django.conf import settings

register = template.Library()


@register.inclusion_tag('otto/analytics.html', takes_context=True)
def analytics(context, next=None):
    # if not getattr(settings, 'OA_ANALYTICS_CREDENTIALS_JSON', None):
    #     raise ImproperlyConfigured('Analytics service account json path missing')
    #
    # if not getattr(settings, 'OA_ANALYTICS_VIEW_ID', None):
    #     raise ImproperlyConfigured('Analytics view id missing')
    #
    # # The scope for the OAuth2 request.
    # SCOPE = 'https://www.googleapis.com/auth/analytics.readonly'
    #
    # # The location of the key file with the key data.
    # KEY_FILEPATH = settings.OA_ANALYTICS_CREDENTIALS_JSON
    #
    # # Load the key file's private data.
    # with open(KEY_FILEPATH) as key_file:
    #     _key_data = json.load(key_file)

    # Construct a credentials objects from the key data and OAuth2 scope.
    # _credentials = SignedJwtAssertionCredentials(_key_data['client_email'], _key_data['private_key'], SCOPE)

    return {
        'token': '_credentials.get_access_token().access_token',
        'view_id': 'settings.OA_ANALYTICS_VIEW_ID'
    }


@register.filter(name='otto_admin_conf')
def otto_admin_conf(name):
    defaults = {
        'OA_COPYRIGHT': '2015 otto.to.it',
        'OA_SUPPORT_EMAIL': 'mail@otto.to.it',
        'OA_POWERED_BY': 'Otto srl',
        'OA_POWERED_BY_URL': 'http://www.otto.to.it',
    }

    return getattr(settings, name, defaults.get(name))



from datetime import datetime
from datetime import time

from django import forms
from django.conf import settings
from django.contrib import admin
from django.utils import timezone
from django.utils.translation import pgettext

from .elegant import ElegantDateWidget


class DateRangeForm(forms.Form):

    def __init__(self, *args, **kwargs):
        """ Automatically generate form fields with dynamic names based on the filtering field name """

        self.field_name = kwargs.pop('field_name', 'date')
        self.title = kwargs.pop('title', 'Date').title()

        super().__init__(*args, **kwargs)

        self.fields[f'{self.field_name}_start'] = forms.DateField(
            widget=ElegantDateWidget(attrs={'placeholder': self.title, 'style': 'width=330px;'}),
            label=pgettext('date', '从'), required=False)

        self.fields[f'{self.field_name}_end'] = forms.DateField(
            widget=ElegantDateWidget(attrs={'placeholder': self.title}),
            label=pgettext('date', '至'), required=False)

    def start_date(self):
        if self.is_valid():
            start = self.cleaned_data.get(f'{self.field_name}_start')

            if start:
                start = datetime.combine(start, time.min)

                if settings.USE_TZ:
                    return timezone.make_aware(start)

            return start

    def end_date(self):
        if self.is_valid():
            end = self.cleaned_data.get(f'{self.field_name}_end')

            if end:
                end = datetime.combine(end, time.max)

                if settings.USE_TZ:
                    return timezone.make_aware(end)

            return end

    class Media:
        css = {'all': ('elegant/css/date_range_filter.css',), }


class DateRangeFilter(admin.FieldListFilter):
    template = 'elegant/date_range_filter.html'
    form = None

    def expected_parameters(self):
        return f'{self.field_path}_start', f'{self.field_path}_end'

    def choices(self, cl):
        return [{'query_string': [], }]

    def get_form(self, request):
        return DateRangeForm(data=request.GET, field_name=self.field_path, title=self.title)

    def queryset(self, request, queryset):
        """
        That's the trick - we create self.form when django tries to get our queryset.
        This allows to create unbount and bound form in the single place.
        """

        self.form = self.get_form(request)

        start_date = self.form.start_date()
        end_date = self.form.end_date()

        if self.form.is_valid() and (start_date or end_date):
            args = self._get_filter_args(start=start_date, end=end_date, )
            return queryset.filter(**args)

    def _get_filter_args(self, start, end):
        filter_args = {}

        if start:
            filter_args[f'{self.field_path}__gte'] = start

        if end:
            filter_args[f'{self.field_path}__lte'] = end

        return filter_args



from django import forms
from django import VERSION
from django.contrib import messages
from django.http import Http404
from django.http import HttpResponse
from django.http import HttpResponseRedirect
from django.urls import re_path as url
from django.utils.safestring import mark_safe
from django.utils.translation import gettext_lazy as _
from django.views import View
from django.views.generic.detail import SingleObjectMixin

from ..components import Dropdown
from ..utils import get_django_model_admin


class AdminRowActionsMixin:
    """ModelAdmin mixin to add row actions just like adding admin actions"""

    # rowactions = []
    _named_row_actions = {}

    @property
    def media(self):
        css = super().media._css  # noqa
        css['all'] = css.get('all', [])
        css['all'].extend(['elegant/css/jquery.dropdown.min.css'])

        js = super().media._js  # noqa
        js.extend(['elegant/js/jquery.dropdown.min.js', ])

        media = forms.Media(css=css, js=js)
        return media

    def get_list_display(self, request):
        self._request = request  # noqa
        list_display = super().get_list_display(request)  # noqa

        if '_row_actions' not in list_display:
            list_display += ('_row_actions',)

        return list_display

    def get_actions_list(self, obj, include_pk=True):

        def to_dict(tool_name):
            return dict(name=tool_name, label=getattr(tool, 'label', tool_name).replace('_', ' ').title())

        items = []

        row_actions = self.get_row_actions(obj)
        url_prefix = '{}/'.format(obj.pk if include_pk else '')

        for tool in row_actions:

            if isinstance(tool, str):  # Just a str naming a callable
                tool_dict = to_dict(tool)

                items.append({
                    'label': tool_dict['label'],
                    'url': f'{url_prefix}rowactions/{tool}/',
                    'method': tool_dict.get('POST', 'GET')
                })

            elif isinstance(tool, dict):  # A parameter dict
                tool['enabled'] = tool.get('enabled', True)

                # If 'action' is specified then use our generic url in preference to 'url' value
                if 'action' in tool:
                    if isinstance(tool['action'], tuple):
                        self._named_row_actions[tool['action'][0]] = tool['action'][1]
                        tool['url'] = f'{url_prefix}rowactions/{tool["action"][0]}/'
                    else:
                        tool['url'] = f'{url_prefix}rowactions/{tool["action"]}/'

                items.append(tool)

        return items

    def _row_actions(self, obj):
        items = self.get_actions_list(obj)

        if items:
            html = Dropdown(label=_('Actions'), items=items, request=getattr(self, '_request')).render()
            return mark_safe(html)

        return ''

    _row_actions.short_description = ''

    if VERSION < (1, 9):
        _row_actions.allow_tags = True

    def get_tool_urls(self):
        """Gets the url patterns that route each tool to a special view"""
        my_urls = [
            url(r'^(?P<pk>[0-9a-f-]+)/rowactions/(?P<tool>\w+)/$',
                self.admin_site.admin_view(ModelToolsView.as_view(model=self.model)))  # noqa
        ]

        return my_urls

    ###################################
    # EXISTING ADMIN METHODS MODIFIED #
    ###################################

    def get_urls(self):
        """Prepends `get_urls` with our own patterns"""
        return self.get_tool_urls() + super().get_urls()  # noqa

    ##################
    # CUSTOM METHODS #
    ##################

    def get_row_actions(self, obj):
        return getattr(self, 'rowactions', False) or []  # noqa

    def get_change_actions(self, request, object_id, form_url):

        # If we're also using django_object_actions
        # then try to reuse row actions as object actions

        change_actions = super().get_change_actions(request, object_id, form_url)  # noqa

        # Make this reuse opt-in
        if getattr(self, 'reuse_row_actions_as_object_actions', False):

            obj = self.model.objects.get(pk=object_id)  # noqa
            row_actions = self.get_actions_list(obj, False) if obj else []

            for row_action in row_actions:
                # Object actions only supports strings as action indentifiers
                if isinstance(row_action, str):
                    change_actions.append(row_action)
                elif isinstance(row_action, dict):
                    if isinstance(row_action['action'], str):
                        change_actions.append(row_action['action'])
                    if isinstance(row_action['action'], tuple):
                        change_actions.append(str(row_action['action'][1]))

        return change_actions


class ModelToolsView(SingleObjectMixin, View):
    """A special view that run the tool's callable"""

    def get(self, request, **kwargs):
        # SingleObjectMixin's `get_object`. Works because the view
        # is instantiated with `model` and the urlpattern has `pk`.

        obj = self.get_object()
        model_admin = get_django_model_admin(obj.__class__)

        if not model_admin:
            raise Http404(f'Can not find ModelAdmin for {obj.__class__}')

        # Look up the action in the following order:
        # 1. in the named_row_actions dict (for lambdas etc)
        # 2. as a method on the model admin
        # 3. as a method on the model
        if kwargs['tool'] in model_admin._named_row_actions:  # noqa
            action_method = model_admin._named_row_actions[kwargs['tool']]  # noqa
            ret = action_method(request=request, queryset=obj)
        elif getattr(model_admin, kwargs['tool'], False):
            action_method = getattr(model_admin, kwargs['tool'])
            # TODO should the signature actually be (obj, request) for consistancy?
            ret = action_method(request=request, queryset=obj)
        elif getattr(obj, kwargs['tool'], False):
            action_method = getattr(obj, kwargs['tool'])
            ret = action_method()
        else:
            raise Http404

        # If the method returns a response use that,
        # otherwise redirect back to the url we were called from
        if isinstance(ret, HttpResponse):
            response = ret
        else:
            back = request.META['HTTP_REFERER']
            response = HttpResponseRedirect(back)

        return response

    # Also allow POST
    post = get

    def message_user(self, request, message):  # noqa
        # Copied from django.contrib.admin.options
        # Included to mimic admin actions
        messages.info(request, message)



import django

try:
    from importlib import import_module
except ImportError:
    pass

try:
    from django.templatetags.future import url
except ImportError:
    from django.template.defaulttags import url

try:
    from django.contrib.contenttypes import admin as ct_admin
except ImportError:
    from django.contrib.contenttypes import generic as ct_admin  # Django < 1.8

if django.VERSION[:2] < (1, 8):
    from django.template import Context

    tpl_context_class = Context
else:
    tpl_context_class = dict



import logging
from functools import wraps

from django import VERSION
from django.db.models.query import QuerySet
from django.urls import reverse

from .constants import CACHE_KEY_PREFIX
from .constants import DEBUG

try:
    from django.contrib.admin.sites import all_sites
except ImportError:
    from django.contrib import admin

    all_sites = [admin.site]

logger = logging.getLogger(__name__)


def django_major_version():
    return VERSION[:2]


def value_by_version(args):
    """
    Return value by version
    Return latest value if version not found
    :param args:
    :return:
    """

    version_map = args_to_dict(args)
    major_version = '.'.join(str(v) for v in django_major_version())
    logger.debug(major_version)

    return version_map.get(major_version, list(version_map.values())[-1])


def args_to_dict(args):
    """
    Convert template tag args to dict
    Format {% elegant_bc 1.5 'x' 1.6 'y' %} to { '1.5': 'x', '1.6': 'y' }
    :param args:
    :return:
    """

    return dict(zip(args[0::2], args[1::2]))


def snake_to_title_case(string: str) -> str:
    return ' '.join(string.split('_')).title()


def get_admin_change_url(obj: object) -> str:
    return reverse(f'admin:{obj._meta.app_label}_{obj._meta.model_name}_change', args=[obj.pk])  # noqa


def format_cache_key(model: str, field: str) -> str:
    return f'{CACHE_KEY_PREFIX}__{model}__{field}'


def log(message: str):  # pragma: no cover
    DEBUG and logger.debug(message)


def inspect(obj: object):  # pragma: no cover
    DEBUG and logger.debug(f'{str(obj): type(obj) - dir(obj)}')


class QuerySetIsh(QuerySet):
    """Takes an instance and mimics it coming from a QuerySet"""

    def __init__(self, instance=None, *args, **kwargs):
        try:
            model = instance._meta.model  # noqa
        except AttributeError:
            # Django 1.5 does this instead, getting the model may be overkill
            # we may be able to throw away all this logic
            model = instance._meta.concrete_model  # noqa

        self._doa_instance = instance
        super().__init__(model, *args, **kwargs)
        self._result_cache = [instance]

    def _clone(self, *args, **kwargs):
        # don't clone me, bro
        return self

    def get(self, *args, **kwargs):
        # Starting in Django 1.7, `QuerySet.get` started slicing to `MAX_GET_RESULTS`,
        # so to avoid messing with `__getslice__`, override `.get`.
        return self._doa_instance


def takes_instance_or_queryset(func):
    """Decorator that makes standard actions compatible"""

    @wraps(func)
    def decorated_function(self, request, queryset):
        # Function follows the prototype documented at:
        # https://docs.djangoproject.com/en/dev/ref/contrib/admin/actions/#writing-action-functions # noqa
        if not isinstance(queryset, QuerySet):
            queryset = QuerySetIsh(queryset)

        return func(self, request, queryset)

    return decorated_function


def get_django_model_admin(model):
    """Search Django ModelAdmin for passed model.

    Returns instance if found, otherwise None.
    """
    for admin_site in all_sites:
        registry = admin_site._registry  # noqa

        if model in registry:
            return registry[model]

    return None



from django import forms
from django.contrib.admin.widgets import AdminDateWidget
from django.contrib.admin.widgets import AdminTimeWidget
from django.forms import Select
from django.forms import Textarea
from django.forms import TextInput
from django.templatetags.static import static
from django.utils.safestring import mark_safe
from django.utils.translation import gettext as _

from .. import utils

django_version = utils.django_major_version()


class NumberInput(TextInput):
    """
    HTML5 Number input
    Left for backwards compatibility
    """

    input_type = 'number'


class HTML5Input(TextInput):
    """
    Supports any HTML5 input
    http://www.w3schools.com/html/html5_form_input_types.asp # noqa
    """

    def __init__(self, attrs=None, input_type=None):
        self.input_type = input_type
        super().__init__(attrs)


#
class LinkedSelect(Select):
    """
    Linked select - Adds link to foreign item, when used with foreign key field
    """

    def __init__(self, attrs=None, choices=()):
        attrs = _make_attrs(attrs, classes='linked-select')
        super().__init__(attrs, choices)


class EnclosedInput(TextInput):
    """
    Widget for bootstrap appended/prepended inputs
    """

    def __init__(self, attrs=None, prepend=None, append=None):
        """
        For prepend, append parameters use string like %, $ or html
        """
        self.prepend = prepend
        self.append = append

        super().__init__(attrs=attrs)

    def enclose_value(self, value):  # noqa
        """
        If value doesn't starts with html open sign "<", enclose in add-on tag
        """
        if value.startswith('<'):
            return value

        if value.startswith('icon-'):
            value = f'<i class="{value}"></i>'

        return f'<span class="add-on">{value}</span>'

    def render(self, name, value, attrs=None, renderer=None):

        if django_version < (2, 0):
            output = super().render(name, value, attrs)
        else:
            output = super().render(name, value, attrs, renderer)

        div_classes = []

        if self.prepend:
            div_classes.append('input-prepend')
            self.prepend = self.enclose_value(self.prepend)
            output = ''.join((self.prepend, output))

        if self.append:
            div_classes.append('input-append')
            self.append = self.enclose_value(self.append)
            output = ''.join((output, self.append))

        return mark_safe('<div class="{}">{}</div>'.format(' '.join(div_classes), output))


class AutosizedTextarea(Textarea):  # noqa
    """
    Autosized Textarea - textarea height dynamically grows based on user input
    """

    def __init__(self, attrs=None):
        new_attrs = _make_attrs(attrs, {'rows': 2}, 'autosize')
        super().__init__(new_attrs)

    @property
    def media(self):
        return forms.Media(js=[static('elegant/js/jquery.autosize-min.js')])

    def render(self, name, value, attrs=None, renderer=None):
        if django_version < (2, 0):
            output = super().render(name, value, attrs)
        else:
            output = super().render(name, value, attrs, renderer)

        output += mark_safe(f"<script type=\"text/javascript\">Elegant.$('#id_{name}').autosize();</script>")

        return output


#
# Original date widgets with addition html
#
class ElegantDateWidget(AdminDateWidget):

    def __init__(self, attrs=None, format=None):  # noqaf
        defaults = {'placeholder': _('Date:')[:-1]}
        new_attrs = _make_attrs(attrs, defaults, 'vDateField input-small')
        super().__init__(attrs=new_attrs, format=format)

    if django_version < (4, 2):
        def render(self, name, value, attrs=None, renderer=None):
            if django_version < (2, 0):
                output = super().render(name, value, attrs)
            else:
                output = super().render(name, value, attrs, renderer)

            return mark_safe(
                f'<div class="input-append elegant-date">{output}<span class="add-on">'
                f'<i class="icon-calendar"></i></span></div>')


class ElegantTimeWidget(AdminTimeWidget):

    def __init__(self, attrs=None, format=None):  # noqa
        defaults = {'placeholder': _('Time:')[:-1]}
        new_attrs = _make_attrs(attrs, defaults, 'vTimeField input-small')
        super().__init__(attrs=new_attrs, format=format)

    if django_version < (4, 2):
        def render(self, name, value, attrs=None, renderer=None):
            if django_version < (2, 0):
                output = super().render(name, value, attrs)
            else:
                output = super().render(name, value, attrs, renderer)

            return mark_safe(f'<div class="input-append elegant-date elegant-time">{output}<span class="add-on">'
                             f'<i class="icon-time"></i></span></div>')


class ElegantSplitDateTimeWidget(forms.SplitDateTimeWidget):
    """
    A SplitDateTime Widget that has some admin-specific styling.
    """

    def __init__(self, attrs=None):
        widgets = [ElegantDateWidget, ElegantTimeWidget]
        forms.MultiWidget.__init__(self, widgets, attrs)

    # if django_version < (1, 11):
    #     def format_output(self, rendered_widgets):
    #         out_tpl = f'<div class="datetime">{rendered_widgets[0]} {rendered_widgets[1]}</div>'
    #         return mark_safe(out_tpl)

    if django_version < (4, 2):
        def render(self, name, value, attrs=None, renderer=None):
            output = super().render(name, value, attrs, renderer)
            return mark_safe('<div class="datetime">%s</div>' % output)


class ElegantProgressWidget(forms.Widget):
    """
    A SplitDateTime Widget that has some admin-specific styling.
    """


def _make_attrs(attrs, defaults=None, classes=None):
    result = defaults.copy() if defaults else {}

    if attrs:
        result.update(attrs)

    if classes:
        result['class'] = ' '.join((classes, result.get('class', '')))

    return result



import re
import warnings

from django import template
from django.contrib import admin
from django.contrib.admin import AdminSite
from django.http import HttpRequest
from django.urls import resolve
from django.urls import reverse

from ..config import get_config

string_types = str,
register = template.Library()
simple_tag = register.simple_tag


@simple_tag(takes_context=True)
def get_menu(context, request):
    """
    :type request: HttpRequest
    """
    if not isinstance(request, HttpRequest):
        return None

    # Try to get app list
    if hasattr(request, 'current_app'):
        # Django 1.8 uses request.current_app instead of context.current_app
        template_response = get_admin_site(request.current_app).index(request)
    else:
        try:
            template_response = get_admin_site(context.current_app).index(request)
        # Django 1.10 removed the current_app parameter for some classes and functions.
        # Check the release notes.
        except AttributeError:
            template_response = get_admin_site(context.request.resolver_match.namespace).index(request)
    try:
        app_list = template_response.context_data['app_list']
    except Exception:

        return

    return Menu(context, request, app_list).get_app_list()


def get_admin_site(current_app):
    """
    Method tries to get actual admin.site class, if any custom admin sites
    were used. Couldn't find any other references to actual class other than
    in func_closer dict in index() func returned by resolver.
    """
    try:
        resolver_match = resolve(reverse(f'{current_app}:index'))
        # Django 1.9 exposes AdminSite instance directly on view function
        if hasattr(resolver_match.func, 'admin_site'):
            return resolver_match.func.admin_site

        for func_closure in resolver_match.func.__closure__:
            if isinstance(func_closure.cell_contents, AdminSite):
                return func_closure.cell_contents
    except:
        pass

    return admin.site


class Menu:
    app_activated = False
    MULTIPLE_MODELS_RE = re.compile(r'([^*]*)[*]')

    def __init__(self, context, request, app_list):
        self.request = request
        self.app_list = app_list

        # Detect current app, if any
        try:
            self.ctx_app = context['app_label'].lower()
        except Exception:
            self.ctx_app = None

        # Get current model plural name, if any
        try:
            self.ctx_model_plural = context['opts'].verbose_name_plural.lower()
        except Exception:
            self.ctx_model_plural = None

        # Flatten all models from native apps
        self.all_models = [model for app in app_list for model in app['models']]

        # Init config variables
        self.init_config()

        super().__init__()

    def init_config(self):
        self.conf_exclude = get_config('MENU_EXCLUDE')
        self.conf_open_first_child = get_config('MENU_OPEN_FIRST_CHILD')
        self.conf_icons = get_config('MENU_ICONS')
        self.conf_menu_order = get_config('MENU_ORDER')
        self.conf_menu = get_config('MENU')

    def get_app_list(self):
        menu = None
        if self.conf_menu:
            menu = self.make_menu(self.conf_menu)
        elif self.conf_menu_order:
            menu = self.make_menu_from_old_format(self.conf_menu_order)
        else:
            menu = self.make_menu_from_native_only()

        # Add icons and match active
        if menu:
            self.activate_menu(menu)

        return menu

    def make_menu(self, config):
        menu = []
        if not isinstance(config, (tuple, list)):
            raise TypeError(f'Django Elegant MENU config parameter must be tuple or list. Got {repr(config)}')
        for app in config:
            app = self.make_app(app)
            if app:
                menu.append(app)

        return menu

    def make_app(self, app_def):
        if isinstance(app_def, dict):
            app = app_def.copy()
        elif isinstance(app_def, string_types):
            if app_def == '-':
                app = self.make_separator()
            else:
                app = self.make_app_from_native(app_def)
        else:
            raise TypeError(f'MENU list item must be string or dict. Got {repr(app_def)}')
        if app:
            return self.process_app(app)

    def process_app(self, app):

        if 'app' in app:
            app = self.process_semi_native_app(app)

        if not app:
            return

        # Process icons
        self.process_icons(app)

        # Ensure required keys for template are set
        self.ensure_app_keys(app)

        # Exclude apps
        if self.app_is_excluded(app):
            return

        # Handle app permissions
        if self.app_is_forbidden(app):
            return

        # Process app models
        self.process_models(app)

        # Set link from child
        models = app.get('models', [])
        if self.conf_open_first_child and models:
            app['orig_url'] = app['url']
            app['url'] = models[0]['url']

        # Process absolute/named/model type urls
        app['url'] = self.process_url(app['url'], app)

        return app

    def app_is_forbidden(self, app):
        return app['permissions'] and not self.user_has_permission(app['permissions'])

    def app_is_excluded(self, app):
        return self.conf_exclude and app['name'] in self.conf_exclude

    def process_icons(self, app):
        """
        If icon key is present but value is '' or None, set empty 'icon-'
        If key not found, try to set icon from SUIT_ICONS
        """
        if 'icon' in app:
            app['icon'] = app['icon'] or 'icon-'

        if self.conf_icons and 'name' in app and app['name'] in self.conf_icons:
            app['icon'] = self.conf_icons[app['name']]

    def process_semi_native_app(self, app):
        """
        Process app defined as { app: 'app' }
        """
        app_from_native = self.make_app_from_native(app['app'])

        if app_from_native:
            del app['app']
            app_from_native.update(app)
            return app_from_native

    def make_menu_from_native_only(self):
        menu = []
        for app in self.app_list:
            app_name = ''
            app_url = app.get('app_url')
            if app_url:
                app_url_parts = app['app_url'].split('/')
                if len(app_url_parts) > 1:
                    app_name = app_url_parts[-2]
            app = self.convert_native_app(app, app_name)
            if app:
                app = self.process_app(app)
            if app:
                menu.append(app)

        return menu

    def make_app_from_native(self, app_name):
        app = self.find_native_app(app_name)
        if app:
            return self.convert_native_app(app, app_name)

    def find_native_app(self, app_name):
        for app in self.app_list:
            _app_name = app.get('app_label', app.get('name'))
            if _app_name.lower() == app_name:
                return app

    def convert_native_app(self, native_app, app_name):
        models = []
        native_models = native_app.get('models', {})

        if native_models:
            for model in native_models:
                models.append(self.convert_native_model(model, app_name))

        # Skip native apps with no models
        if not models:
            return

        return {
            'label': native_app['name'],
            'url': native_app['app_url'],
            'models': models,
            'name': app_name
        }

    def make_separator(self):
        return {'separator': True}

    def process_models(self, app):
        models = []
        models_def = app.get('models', [])

        for model_def in models_def:
            # multiple models may be returned
            models += self.make_models(model_def, app['name'])

        app['models'] = models

    def make_models(self, model_def, app_name):
        if not isinstance(model_def, string_types):
            model = self.make_model(model_def, app_name)
            return [model] if model else []

        match = self.MULTIPLE_MODELS_RE.match(model_def)

        if not match:
            model = self.make_model(model_def, app_name)
            return [model] if model else []

        prefix = match.group(1)
        prefix = self.get_model_name(app_name, prefix)

        return [
            m for m in [
                self.convert_native_model(native_model, app_name)
                for native_model in self.all_models
                if self.get_native_model_name(native_model).startswith(prefix)
            ] if m
        ]

    def make_model(self, model_def, app_name):
        if isinstance(model_def, dict):
            model = model_def.copy()
        elif isinstance(model_def, string_types):
            model = self.make_model_from_native(model_def, app_name)
        else:
            raise TypeError(f'MENU list item must be string or dict. Got {repr(model_def)}')

        if model:
            return self.process_model(model, app_name)

    def make_model_from_native(self, model_name, app_name):
        model = self.find_native_model(model_name, app_name)
        if model:
            return self.convert_native_model(model, app_name)

    def find_native_model(self, model_name, app_name):
        model_name = self.get_model_name(app_name, model_name)
        for native_model in self.all_models:
            if model_name == self.get_native_model_name(native_model):
                return native_model

    def model_is_excluded(self, model_name):
        return self.conf_exclude and model_name in self.conf_exclude

    def get_model_name(self, app_name, model_name):
        if app_name:
            app_name = app_name.lower()

        if model_name:
            model_name = model_name.lower()

        if '.' not in model_name:
            model_name = f'{app_name}.{model_name}'

        return model_name

    def get_native_model_name(self, model):
        """
        Get model name by its last part of url
        """
        url_parts = self.get_native_model_url(model).rstrip('/').split('/')
        root_url_parts = reverse('admin:index').rstrip('/').split('/')
        return '.'.join(url_parts[len(root_url_parts):][:2])

    def convert_native_model(self, model, app_name):
        return {
            'label': model['name'],
            'url': self.get_native_model_url(model),
            'name': self.get_native_model_name(model),
            'app': app_name,
            'perms': model.get('perms', None),
            'add_url': model.get('add_url', None),
        }

    def get_native_model_url(self, model):
        return model.get('admin_url') or model.get('add_url', '')

    def process_model(self, model, app_name):
        if 'model' in model:
            model = self.process_semi_native_model(model, app_name)

        if model:
            self.ensure_model_keys(model)

            if 'app' in model and 'name' in model:
                model_name = self.get_model_name(model['app'], model['name'])

                if self.model_is_excluded(model_name):
                    return

            # Handle model permissions
            if self.model_is_forbidden(model):
                return

            # Detect if named url and convert it to absolute
            model['url'] = self.process_url(model['url'])

            return model

    def model_is_forbidden(self, model):
        return model['permissions'] and not self.user_has_permission(model['permissions'])

    def process_semi_native_model(self, model, app_name):
        """
        Process app defined as { model: 'model' }
        """
        model_from_native = self.make_model_from_native(model['model'], app_name)
        if model_from_native:
            del model['model']
            model_from_native.update(model)
            return model_from_native

    def ensure_app_keys(self, app):
        keys = [
            'label', 'url', 'icon', 'permissions', 'name', 'is_active', 'blank'
        ]
        self.fill_keys(app, keys)

    def ensure_model_keys(self, model):
        keys = ['label', 'url', 'permissions', 'is_active', 'blank']
        self.fill_keys(model, keys)

    def fill_keys(self, dict, keys):
        for key in keys:
            if key not in dict:
                dict[key] = None

    def user_has_permission(self, perms):
        perms = perms if isinstance(perms, (list, tuple)) else (perms,)
        return self.request.user.has_perms(perms)

    def activate_menu(self, menu):
        for app in menu:

            # Make 'model' key as 'models' to unite activation logic
            if 'model' in app and not app['models']:
                app['models'] = [app['model']]

            # Activate models
            if app['models']:
                self.activate_models(app)

            # Mark as active by url match
            if not self.app_activated and (self.request.path == app['url'] or self.request.path == app.get('orig_url')):
                app['is_active'] = self.app_activated = True

        if not self.app_activated:
            self.activate_menu_by_url(menu)

        # Last chance, try to activate by name
        if not self.app_activated:
            for app in menu:
                self.activate_models(app, match_by_name=True)

    def activate_models(self, app, match_by_name=False):
        for model in app['models']:
            if not match_by_name:
                # Mark as active by url or model plural name match
                model['is_active'] = self.request.path == model['url']
            else:
                model['is_active'] = self.ctx_model_plural == model[
                    'label'].lower()

            # Mark parent as active too
            if model['is_active'] and not self.app_activated:
                app['is_active'] = self.app_activated = True

    def activate_menu_by_url(self, menu):
        """
        If no active app/model is found in good/correct way, try to match
        by simple "startswith" in request path. Some apps doesn't provide
        nice app_label (django-filer for ex.) therefore this is the only way
        """
        for app in menu:
            for model in app['models']:
                if model['url'] and self.request.path.startswith(model['url']):
                    model['is_active'] = True
                    app['is_active'] = self.app_activated = True
                    break
            if self.app_activated:
                break

        # If still no active app found, match by app original url if any
        if not self.app_activated:
            for app in menu:
                orig_url = app.get('orig_url')
                if orig_url and self.request.path.startswith(orig_url):
                    app['is_active'] = self.app_activated = True

    def process_url(self, url, app=None):
        """
        Try to guess if it is absolute url or named
        """
        if url is None:
            return ''

        if not url or '/' in url:
            return url

        # Model link, ex: 'auth.user'
        if '.' in url:
            url_parts = url.split('.')
            model = self.make_model_from_native(url_parts[1], url_parts[0])
            if model:
                if app:
                    app['model'] = model
                return model['url']

        # Try to resolve as named url, ex: 'admin:index'
        try:
            return reverse(url)
        except:
            return url

    def make_menu_from_old_format(self, conf_order):
        import sys

        if 'test' not in sys.argv:
            warnings.warn(
                'Django Elegant "MENU_ORDER" setting is deprecated. Use new "MENU"'
                ' key instead, see Documentation for new syntax.',
                DeprecationWarning)

        new_conf = []

        for order in conf_order:
            new_app = {}

            if isinstance(order, (tuple, list)):
                app_name = order[0]
                models_order = order[1] if len(order) > 1 else None

                if isinstance(app_name, string_types):
                    new_app['app'] = app_name
                elif isinstance(app_name, (tuple, list)):
                    mapping = ('label', 'url', 'icon', 'permissions')
                    for i, val in enumerate(app_name):
                        new_app[mapping[i]] = val

                if models_order and isinstance(models_order, (tuple, list)):
                    models = []
                    for model in models_order:
                        if isinstance(model, string_types):
                            models.append({'model': model})
                        elif isinstance(model, (list, tuple)):
                            mapping = ('label', 'url', 'permissions')
                            new_model = {}
                            for i, val in enumerate(model):
                                new_model[mapping[i]] = val
                            models.append(new_model)

                    new_app['models'] = models

            if new_app:
                new_conf.append(new_app)

        return self.make_menu(new_conf)



import inspect
import logging
from copy import copy
from urllib.parse import parse_qs

from django import template
from django.contrib.admin.templatetags.admin_list import result_list
from django.contrib.admin.views.main import ALL_VAR
from django.contrib.admin.views.main import PAGE_VAR
from django.template.loader import get_template
from django.utils.html import escape
from django.utils.safestring import mark_safe

from ..compat import tpl_context_class
from elegant import utils

logger = logging.getLogger(__name__)
register = template.Library()

DOT = '.'
django_version = utils.django_major_version()


@register.simple_tag
def paginator_number(cl, i):
    """
    Generates an individual page index link in a paginated list.
    """

    if i == DOT:
        return mark_safe('<li class="disabled"><a href="#" onclick="return false;">...</a></li>')

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    # < 3.2: 0,1,2,3,4,5,6
    # >= 3.2: 1,2,3,4,5,6,7
    p = (i, i + 1)[django_version < (3, 2)]

    if i == cl.page_num:
        return mark_safe(f'<li class="active"><a href="">{p}</a></li>')

    link = escape(cl.get_query_string({PAGE_VAR: i}))
    attr = i >= cl.paginator.num_pages and 'class="end"' or ''
    return mark_safe(f'<li><a href="{link}" {attr}>{p}</a></li>')


@register.simple_tag
def paginator_info(cl):
    paginator = cl.paginator

    # If we show all rows of list (without pagination)
    if cl.show_all and cl.can_show_all:
        entries_from = 1 if paginator.count > 0 else 0
        entries_to = paginator.count
        print(f'entries_from paginator.count => {entries_from}')

    else:
        if django_version < (3, 2):
            entries_from = ((paginator.per_page * cl.page_num) + 1) if paginator.count > 0 else 0
        else:
            entries_from = ((paginator.per_page * (cl.page_num - 1)) + 1) if paginator.count > 0 else 0

        entries_to = entries_from - 1 + paginator.per_page

        if paginator.count < entries_to:
            entries_to = paginator.count

    return f'{entries_from} - {entries_to}'


@register.inclusion_tag('admin/pagination.html')
def pagination(cl):
    """
    Generates the series of links to the pages in a paginated list.
    """
    paginator, page_num = cl.paginator, cl.page_num
    pagination_required = (not cl.show_all or not cl.can_show_all) and cl.multi_page

    if not pagination_required:
        page_range = []
    else:
        on_each_side = 3
        on_ends = 2

        # If there are 10 or fewer pages, display links to every page.
        # Otherwise, do some fancy
        if paginator.num_pages <= 8:
            page_range = range(1, paginator.num_pages + 1)
        else:
            # Insert "smart" pagination links, so that there are always ON_ENDS
            # links at either end of the list of pages, and there are always
            # ON_EACH_SIDE links at either end of the "current page" link.
            page_range = []

            # 大于前后预留数和, 则中间加 ...
            if page_num > (on_each_side + on_ends):
                page_range.extend(list(range(1, on_each_side + 1)))
                page_range.append(DOT)
                page_range.extend(range(page_num - on_each_side, page_num + 1))
            else:
                page_range.extend(range(1, page_num + 1))

            if page_num < (paginator.num_pages - on_each_side - on_ends - 1):
                page_range.extend(range(page_num + 1, page_num + on_each_side + 1))
                page_range.append(DOT)
                page_range.extend(range(paginator.num_pages - on_ends, paginator.num_pages + 1))
            else:
                page_range.extend(range(page_num + 1, paginator.num_pages + 1))

    need_show_all_link = cl.can_show_all and not cl.show_all and cl.multi_page

    return {
        'cl': cl,
        'pagination_required': pagination_required,
        'show_all_url': need_show_all_link and cl.get_query_string({ALL_VAR: ''}),
        'page_range': page_range,
        'ALL_VAR': ALL_VAR,
        '1': 1,
    }


@register.simple_tag
def elegant_list_filter_select(cl, spec):
    tpl = get_template(spec.template)
    choices = list(spec.choices(cl))
    field_key = spec.field_path if hasattr(spec, 'field_path') else spec.parameter_name
    matched_key = field_key

    for choice in choices:
        query_string = choice['query_string'][1:]
        query_parts = parse_qs(query_string)

        value = ''
        matches = {}

        for key in query_parts.keys():
            if key == field_key:
                value = query_parts[key][0]
                matched_key = key
            elif key.startswith(field_key + '__') or '__' + field_key + '__' in key:
                value = query_parts[key][0]
                matched_key = key

            if value:
                matches[matched_key] = value

        # Iterate matches, use first as actual values, additional for hidden
        i = 0

        for key, value in matches.items():
            if i == 0:
                choice['name'] = key
                choice['val'] = value
            else:
                choice['additional'] = f'{key}={value}'

            i += 1

    return tpl.render(
        tpl_context_class({
            'field_name': field_key,
            'title': spec.title,
            'choices': choices,
            'spec': spec,
        }))


@register.filter
def headers_handler(result_headers, cl):
    """
    Adds field name to css class, so we can style specific columns
    """
    # field = cl.list_display.get()
    attrib_key = 'class_attrib'

    for i, header in enumerate(result_headers):
        field_name = cl.list_display[i]

        if field_name == 'action_checkbox':
            continue

        if not attrib_key in header:
            header[attrib_key] = mark_safe(' class=""')

        pattern = 'class="'

        if pattern in header[attrib_key]:
            replacement = f'{pattern}{field_name}-column '
            header[attrib_key] = mark_safe(header[attrib_key].replace(pattern, replacement))

    return result_headers


def dict_to_attrs(attrs):
    return mark_safe(' ' + ' '.join([f'{k}="{v}"' for k, v in attrs.items()]))


@register.inclusion_tag('admin/change_list_results.html', takes_context=True)
def result_list_with_context(context, cl):
    """
    Wraps Djangos default result_list to ammend the context with the request.

    This gives us access to the request in change_list_results.
    """
    res = result_list(cl)
    res['request'] = context['request']

    return res


@register.simple_tag(takes_context=True)
def result_row_attrs(context, cl, row_index):
    """
    Returns row attributes based on object instance
    """
    row_index -= 1
    attrs = {'class': 'row1' if row_index % 2 == 0 else 'row2'}
    elegant_row_attributes = getattr(cl.model_admin, 'elegant_row_attributes', None)

    if not elegant_row_attributes:
        return dict_to_attrs(attrs)

    instance = cl.result_list[row_index]

    # Backwards compatibility for elegant_row_attributes without request argument
    # todo args = getargspec(elegant_row_attributes)
    args = inspect.getfullargspec(elegant_row_attributes)

    if 'request' in args[0]:
        new_attrs = elegant_row_attributes(instance, context['request'])
    else:
        new_attrs = elegant_row_attributes(instance)

    if not new_attrs:
        return dict_to_attrs(attrs)

    # Validate
    if not isinstance(new_attrs, dict):
        raise TypeError(f'"elegant_row_attributes" must return dict. Got: {new_attrs.__class__.__name__}: {new_attrs}')

    # Merge 'class' attribute
    if 'class' in new_attrs:
        attrs['class'] += ' ' + new_attrs.pop('class')

    attrs.update(new_attrs)
    return dict_to_attrs(attrs)


@register.filter
def cells_handler(results, cl):
    """
    Changes result cell attributes based on object instance and field name
    """
    elegant_cell_attributes = getattr(cl.model_admin, 'elegant_cell_attributes', None)
    if not elegant_cell_attributes:
        return results

    class_pattern = 'class="'
    td_pattern = '<td'
    th_pattern = '<th'

    for row, result in enumerate(results):
        instance = cl.result_list[row]
        for col, item in enumerate(result):
            field_name = cl.list_display[col]
            attrs = copy(elegant_cell_attributes(instance, field_name))
            if not attrs:
                continue

            # Validate
            if not isinstance(attrs, dict):
                raise TypeError(f'"elegant_cell_attributes" must return dict. Got: {attrs.__class__.__name__}: {attrs}')

            # Merge 'class' attribute
            if class_pattern in item.split('>')[0] and 'class' in attrs:
                css_class = attrs.pop('class')
                replacement = f'{class_pattern}{css_class} '
                result[col] = mark_safe(item.replace(class_pattern, replacement))

            # Add rest of attributes if any left
            if attrs:
                cell_pattern = td_pattern if item.startswith(td_pattern) else th_pattern
                result[col] = mark_safe(result[col].replace(cell_pattern, td_pattern + dict_to_attrs(attrs)))

    return results
